/*
 * Copyright 2016 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import QtQuick 2.4
import QtQuick.Layouts 1.1
import QtQuick.Window 2.2
import Ubuntu.Components 1.3
import Ubuntu.Components.ListItems 1.3 as ListItems
import "../Components"

UbuntuShape {
    id: root
    objectName: "menu"
    backgroundColor: theme.palette.normal.overlay

    property alias unityMenuModel: repeater.model

    function show() {
        visible = true;
        focusScope.forceActiveFocus();
    }

    function hide() {
        visible = false;
        d.currentItem = null;
    }

    function select(index) {
        d.select(index)
    }

    function reset() {
        d.currentItem = null;
        dismiss();
    }

    function dismiss() {
        d.dismissAll();
    }

    implicitWidth: focusScope.width
    implicitHeight: focusScope.height

    MenuNavigator {
        id: d
        objectName: "d"
        itemView: repeater

        property Item currentItem: null
        property Item hoveredItem: null
        readonly property int currentIndex: currentItem ? currentItem.__ownIndex : -1

        property real __minimumWidth: units.gu(20)
        property real __maximumWidth: Screen.width * 0.7
        property real __minimumHeight: units.gu(2)
        property real __maximumHeight: Screen.height * 0.7

        signal dismissAll()

        onCurrentItemChanged: {
            if (currentItem) {
                currentItem.item.forceActiveFocus();
            } else {
                hoveredItem = null;
            }
        }

        onSelect: {
            currentItem = repeater.itemAt(index);
            if (currentItem) {
                if (currentItem.y < listView.contentY) {
                    listView.contentY = currentItem.y;
                } else if (currentItem.y + currentItem.height > listView.contentY + listView.height) {
                    listView.contentY = currentItem.y + currentItem.height - listView.height;
                }
            }
        }
    }

    MouseArea {
        // Eat events.
        anchors.fill: parent
    }

    Item {
        id: focusScope
        width: container.width
        height: container.height
        focus: visible

        Keys.onUpPressed: d.selectPrevious(d.currentIndex)
        Keys.onDownPressed: d.selectNext(d.currentIndex)
        Keys.onRightPressed: {
            // Don't let right keypresses fall through if the current item has a visible popup.
            if (!d.currentItem || !d.currentItem.popup || !d.currentItem.popup.visible) {
                event.accepted = false;
            }
        }

        ColumnLayout {
            id: container
            objectName: "container"

            height: MathUtils.clamp(listView.contentHeight, d.__minimumHeight, d.__maximumHeight)
            width: menuColumn.width
            spacing: 0

            // Header - scroll up
            Item {
                Layout.fillWidth: true
                height: units.gu(3)
                visible: listView.contentHeight > root.height
                enabled: !listView.atYBeginning
                z: 1

                Rectangle {
                    color: enabled ? theme.palette.normal.overlayText :
                        theme.palette.disabled.overlayText
                    height: units.dp(1)
                    anchors {
                        bottom: parent.bottom
                        left: parent.left
                        right: parent.right
                    }
                }

                Icon {
                    anchors.centerIn: parent
                    width: units.gu(2)
                    height: units.gu(2)
                    name: "up"
                    color: enabled ? theme.palette.normal.overlayText :
                                     theme.palette.disabled.overlayText
                }

                MouseArea {
                    anchors.fill: parent
                    onPressed: {
                        var item = menuColumn.childAt(0, listView.contentY);
                        if (item) {
                            var previousItem = item;
                            do {
                                previousItem = repeater.itemAt(previousItem.__ownIndex-1);
                                if (!previousItem) {
                                    listView.contentY = 0;
                                    return;
                                }
                            } while (previousItem.__isSeparator);

                            listView.contentY = previousItem.y
                        }
                    }
                }
            }

            // Menu Items
            Flickable {
                id: listView
                clip: interactive

                Layout.fillHeight: true
                Layout.fillWidth: true
                contentHeight: menuColumn.height
                interactive: height < contentHeight

                MouseArea {
                    anchors.fill: parent
                    hoverEnabled: true
                    propagateComposedEvents: true // propogate events so we send clicks to children.
                    z: 1 // on top so we override any other hovers
                    onEntered: updateCurrentItemFromPosition(Qt.point(mouseX, mouseY))
                    onPositionChanged: updateCurrentItemFromPosition(Qt.point(mouse.x, mouse.y))

                    function updateCurrentItemFromPosition(point) {
                        var pos = mapToItem(listView.contentItem, point.x, point.y);

                        if (!d.hoveredItem || !d.currentItem ||
                                !d.hoveredItem.contains(Qt.point(pos.x - d.currentItem.x, pos.y - d.currentItem.y))) {
                            d.hoveredItem = menuColumn.childAt(pos.x, pos.y)
                            if (!d.hoveredItem || !d.hoveredItem.enabled)
                                return false;
                            d.currentItem = d.hoveredItem;
                        }
                        return true;
                    }
                }

                ActionContext {
                    id: menuBarContext
                    objectName: "menuContext"
                    active: {
                        if (!root.visible) return false;
                        if (d.currentItem && d.currentItem.popup && d.currentItem.popup.visible) {
                            return false;
                        }
                        return true;
                    }
                }

                ColumnLayout {
                    id: menuColumn
                    spacing: 0

                    width: MathUtils.clamp(implicitWidth, d.__minimumWidth, d.__maximumWidth)

                    Repeater {
                        id: repeater

                        Loader {
                            id: loader
                            objectName: root.objectName + "-item" + __ownIndex

                            property int __ownIndex: index
                            property bool __isSeparator: model.isSeparator

                            enabled: __isSeparator ? false : model.sensitive

                            sourceComponent: {
                                if (model.isSeparator) {
                                    return separatorComponent;
                                }
                                return menuItemComponent;
                            }

                            property Item popup: null

                            Layout.fillWidth: true

                            Component {
                                id: menuItemComponent
                                MenuItem {
                                    id: menuItem
                                    menuData: model
                                    objectName: loader.objectName + "-actionItem"

                                    width: MathUtils.clamp(implicitWidth, d.__minimumWidth, d.__maximumWidth)

                                    action.onTriggered: {
                                        d.currentItem = loader;

                                        if (hasSubmenu) {
                                            if (!popup) {
                                                var model = root.unityMenuModel.submenu(__ownIndex);
                                                popup = submenuComponent.createObject(focusScope, {
                                                                                          objectName: loader.objectName + "-",
                                                                                          unityMenuModel: model,
                                                                                          x: Qt.binding(function() { return root.width }),
                                                                                          y: Qt.binding(function() {
                                                                                              var dummy = listView.contentY; // force a recalc on contentY change.
                                                                                              return mapToItem(container, 0, y).y;
                                                                                          })
                                                                                      });
                                            } else if (popup) {
                                                popup.visible = true;
                                            }
                                            popup.retreat.connect(function() {
                                                popup.destroy();
                                                popup = null;
                                                menuItem.forceActiveFocus();
                                            })
                                        } else {
                                            root.unityMenuModel.activate(__ownIndex);
                                        }
                                    }

                                    Connections {
                                        target: d
                                        onCurrentIndexChanged: {
                                            if (popup && d.currentIndex != __ownIndex) {
                                                popup.visible = false;
                                            }
                                        }
                                        onDismissAll: {
                                            if (popup) {
                                                popup.destroy();
                                                popup = null;
                                            }
                                        }
                                    }
                                }
                            }

                            Component {
                                id: separatorComponent
                                ListItems.ThinDivider {
                                    objectName: loader.objectName + "-separator"
                                    implicitHeight: units.dp(2)
                                }
                            }
                        }

                    }
                }

                // Highlight
                Rectangle {
                    color: "transparent"
                    border.width: units.dp(1)
                    border.color: UbuntuColors.orange
                    z: 1

                    width: listView.width
                    height:  d.currentItem ? d.currentItem.height : 0
                    y:  d.currentItem ? d.currentItem.y : 0
                    visible: d.currentItem
                }

            } // Flickable

            // Header - scroll down
            Item {
                Layout.fillWidth: true
                height: units.gu(3)
                visible: listView.contentHeight > root.height
                enabled: !listView.atYEnd
                z: 1

                Rectangle {
                    color: enabled ? theme.palette.normal.overlayText :
                                     theme.palette.disabled.overlayText
                    height: units.dp(1)
                    anchors {
                        top: parent.top
                        left: parent.left
                        right: parent.right
                    }
                }

                Icon {
                    anchors.centerIn: parent
                    width: units.gu(2)
                    height: units.gu(2)
                    name: "down"
                    color: enabled ? theme.palette.normal.overlayText :
                                     theme.palette.disabled.overlayText
                }

                MouseArea {
                    anchors.fill: parent
                    onPressed: {
                        var item = menuColumn.childAt(0, listView.contentY + listView.height);
                        if (item) {
                            var nextItem = item;
                            do {
                                nextItem = repeater.itemAt(nextItem.__ownIndex+1);
                                if (!nextItem) {
                                    listView.contentY = listView.contentHeight - listView.height;
                                    return;
                                }
                            } while (nextItem.__isSeparator);

                            listView.contentY = nextItem.y - listView.height
                        }
                    }
                }
            }
        } // Column

        Component {
            id: submenuComponent
            Loader {
                id: submenuLoader
                source: "MenuPopup.qml"

                property var unityMenuModel: null
                signal retreat()

                Binding {
                    target: item
                    property: "unityMenuModel"
                    value: submenuLoader.unityMenuModel
                }

                Binding {
                    target: item
                    property: "objectName"
                    value: submenuLoader.objectName + "menu"
                }

                Keys.onLeftPressed: retreat()

                Component.onCompleted: item.select(0);
                onVisibleChanged: if (visible) { item.select(0); }
            }
        }
    }
}
